-------------------------------------------------------------
-- MSS copyright 2004
--	Filename:  COM5101A.VHD
-- Author: Alain Zarembowitch / MSS
--	Version: 3
--	Date last modified: 7-27-04
-- Inheritance: 	none
--
-- description:  CPLD design for signal conditioning, reformatting, 
-- reclocking, routing between 2 40-pin comblock interfaces and 
-- a DB25 connector with RS422/RS232 interfaces.
-- -A firmware option.
-- 
-- Design Note:
-- 
-- User notes:
--
-- Rev 2: 4-3-04 AZ
-- Reinstated serial link for hardware rev2
--
-- Rev 3: 7-27-04 AZ
-- Added connection matrices A4/A5: 
-- Two signal input to ComBlock assembly:
-- clock towards data source, data from source.
-- Flow control looped back internally
-- A4: data is sent at rising edge of clock, read at falling edge of clock
-- A5: data is sent at falling edge of clock, read at rising edge of clock
---------------------------------------------------------------
library IEEE;
use IEEE.STD_LOGIC_1164.ALL;
use IEEE.STD_LOGIC_ARITH.ALL;
use IEEE.STD_LOGIC_UNSIGNED.ALL;

entity com5101A is
    Port ( 
	 
		--GLOBAL CLOCKS
	   CLK_IN1 : in std_logic;		-- master clock for this CPLD, internal to the COM5101module
	   CLK_IN2 : in std_logic;	-- master clock for this CPLD, external to the module
	 
		--// Left connector J2, 40 pin female. 
		-- Pin 1 is reserved for input clock CLK_IN2
		-- Pins 10,20,30,40 reserved for Ground
		-- Pins 37,38,39 reserved for JTAG
		-- Pins 35-36 reserved for ComBlock Monitoring and Control.
		J2: inout std_logic_vector(29 downto 2);

		--// Bottom connectors J6 , 40 pin male. 
		-- Pin 1 is reserved to forward clock to following module.
		-- Pins 10,20,30,40 reserved for Ground
		-- Pins 37,38,39 reserved for JTAG
		-- Pins 35-36 reserved for ComBlock Monitoring and Control.
		J6: inout std_logic_vector(31 downto 1);

		--// Serial ports interface with ComBlock J2/J6 connectors
		-- signals are tied together for left and bottom connectors.
		RX_LB: in std_logic;
		TX_LB: out std_logic;

		--// DB25 interface
      RS232_RX : in std_logic;
      RS232_TX : out std_logic;
      RS422_TX : out std_logic_vector(4 downto 1);
      RS422_RX : in std_logic_vector(4 downto 1);

		--// Configuration selection jumpers
		-- User can select up to 16 configurations
		CPLD_CNTRL: in std_logic_vector(3 downto 0)
);
end entity;

architecture Behavioral of com5101A is
--------------------------------------------------------
--      COMPONENTS
--------------------------------------------------------
component BUFG
      port (I : in STD_LOGIC; O : out std_logic);
end component;
--------------------------------------------------------
--     SIGNALS
--------------------------------------------------------
-- Suffix _D indicates a one CLK delayed version of the net with the same name
-- Suffix _E indicates an extended precision version of the net with the same name
-- Suffix _N indicates an inverted version of the net with the same name

--// CLOCKS, RESETS
signal CLK_IN1G: 	std_logic;
signal CLK_IN2G: 	std_logic;

signal RS422_RX_D: std_logic_vector(4 downto 1);
signal RS422_RX_D2: std_logic_vector(1 downto 1);
signal RS422_TX4_B: std_logic;
signal BIT_COUNT: std_logic_vector(2 downto 0);
signal DATA_BYTE: std_logic_vector(6 downto 0);
signal DATA_BYTE_OUT: std_logic_vector(7 downto 0);
signal TRIGGER_OUT: std_logic;
signal SAMPLE_CLK_OUT: std_logic;
signal J6_A0_A1: std_logic_vector(5 downto 2);

--// to RS422
signal COUNTER: std_logic_vector(1 downto 0);
signal  RS422_TEMP: std_logic_vector(5 downto 2);
signal RS422_TX1_A: std_logic;
signal RS422_TX2_A: std_logic;
signal RS422_TX3_A: std_logic;
signal RS422_TX4_A: std_logic;

signal J6_13_D: std_logic;
signal J6_13_D2: std_logic;
signal J6_13_D3: std_logic;
--------------------------------------------------------
--      IMPLEMENTATION
--------------------------------------------------------
begin

-- Generate global clocks
BUFG_001: BUFG  port map (I=> CLK_IN1, O=> CLK_IN1G);
BUFG_002: BUFG  port map (I=> CLK_IN2, O=> CLK_IN2G);

--// OPTION -A FIRMWARE --------------------------

--// CLK_OUT generation
-- Invert clock for best timing.
J6(1) <= not CLK_IN1G;

--// RS232: connect to J2/J6
TX_LB <= RS232_RX;
RS232_TX <= RX_LB;

--// Reformat ComBlock (via J2) to RS422 signals
-- Reclock all inputs
RECLOCK_001: process(CLK_IN2G)
begin
	if rising_edge(CLK_IN2G) then
		-- extend sampling pulse 
		if(J2(2) = '1') then
			-- new incoming sample
			-- Three cases: 
			if(COUNTER /= 0) then
				-- transmitting last sample on slower RS422. Save
				-- incoming sample in temporary storage.
				RS422_TEMP(5 downto 2) <= J2(5 downto 2);
				COUNTER <= COUNTER - 1;
			elsif(RS422_TEMP(2) = '0') then
				-- no sample in temp storage. forward new sample to RS422
				-- re-arm counter
				COUNTER <= "11";
				RS422_TX2_A <= J2(3);
				RS422_TX3_A <= J2(4);
				RS422_TX4_A <= J2(5);
			else
				-- sample in temp storage has precedence. 
				-- save new sample in temp storage
				-- re-arm counter
				COUNTER <= "11";
				RS422_TEMP(5 downto 2) <= J2(5 downto 2);
				RS422_TX2_A <= RS422_TEMP(3);
				RS422_TX3_A <= RS422_TEMP(4);
				RS422_TX4_A <= RS422_TEMP(5);
			end if;
	 	else
			-- no new sample. 
			-- Check if sample is waiting in temp storage
			-- re-arm counter
			if(COUNTER /= 0) then
				-- transmitting last sample on slower RS422. 
				-- just decrement counter
				COUNTER <= COUNTER - 1;
			elsif(RS422_TEMP(2) = '1') then
				-- not currently transmitting on RS422.
				-- Move samples in temporary storage to RS422
				COUNTER <= "11";
				RS422_TEMP(2) <= '0';	-- clear temp storage flag
				-- forward temp storage to RS422
				RS422_TX2_A <= RS422_TEMP(3);
				RS422_TX3_A <= RS422_TEMP(4);
				RS422_TX4_A <= RS422_TEMP(5);
			end if;
	 	end if;

  	end if;
end process;

-- clock is 50 ns wide, in the middle of 100ns data perid.
RS422_TX1_A <= '1' when ((COUNTER = 2) or (COUNTER = 1)) else '0';


-- RS422 signal clocked with CLK_IN1G
-- Aggregate all requests until the next rising edge of the RS422_RX1 sampling clock
RECLOCK_003: process(CLK_IN1G)
begin
	if rising_edge(CLK_IN1G) then
		if (CPLD_CNTRL = 1) and (J6(13) = '1') then
			-- A1 connection matrix
			-- sample request received
			RS422_TX4_B <= '1';	
		elsif(RS422_RX(1) = '1') and (RS422_RX_D(1) = '0') then
			-- keep request until after the next rising edge of the RS422_RX1 sampling clock
		 	-- when it is sure  to have been read by the user 
			RS422_TX4_B <= '0';	
	  	end if;
  	end if;
end process;

--// RS422 TX OUTPUT MUX ---------------------
RS422_TX_GEN: process(CPLD_CNTRL, J2, RS422_TX1_A, RS422_TX2_A, RS422_TX3_A)
begin
	case(CPLD_CNTRL) is
		when "0000" => RS422_TX(3 downto 1) <= RS422_TX3_A & RS422_TX2_A & RS422_TX1_A;
		when "0001" => RS422_TX(3 downto 1) <= RS422_TX3_A & RS422_TX2_A & RS422_TX1_A;
		when "0011" => RS422_TX(3 downto 1) <= J2(4) & J2(3) & J2(2);
		when "0100" => RS422_TX(3 downto 1) <= "ZZ" & (J6_13_D or J6_13_D2);
		when "0101" => RS422_TX(3 downto 1) <= "ZZ" & (not(J6_13_D or J6_13_D2));
		when others => RS422_TX(3 downto 1) <= (others => 'Z');
	end case;
end process;


-- Signal on RS422_TX(4) can stem from sources with different clocks 
-- which cannot be generated within the same process. Consolidate here by ORing.
RS422_TX4_GEN: process(CPLD_CNTRL, RS422_TX4_A, RS422_TX4_B, J2)
begin
	case(CPLD_CNTRL) is
		when "0000" => RS422_TX(4) <= RS422_TX4_A;
		when "0001" => RS422_TX(4) <= RS422_TX4_B;
		when "0011" => RS422_TX(4) <= J2(5);
		when others => RS422_TX(4) <= 'Z';
	end case;
end process;


-- Generate RS422 to ComBlock (via J6) signals
-- Double reclock with 40 MHz internal signal
RECLOCK_004: process(CLK_IN1G, RS422_RX)
begin
	if rising_edge(CLK_IN1G) then
		RS422_RX_D <= RS422_RX;
		RS422_RX_D2(1) <= RS422_RX_D(1);
  	end if;
end process;

-- If RS422_RX(1) goes from zero to one, generate pulse on J6(2);
CONVERSION_001: process(CLK_IN1G)
begin
	if rising_edge(CLK_IN1G) then
		if(RS422_RX_D(1) = '1') and (RS422_RX_D2(1) = '0') then
			J6_A0_A1(2) <= '1';
			J6_A0_A1(3) <= RS422_RX_D(2);
			J6_A0_A1(4) <= RS422_RX_D(3);
			J6_A0_A1(5) <= RS422_RX_D(4);
		else
			J6_A0_A1(2) <= '0';
		end if;
  	end if;
end process; 

--// Connection matrix A2 -------------------------
-- Bit counter to perform serial to 8-bit parallel conversion
BIT_COUNT_GEN_001: process(CLK_IN1G)
begin
	if rising_edge(CLK_IN1G) then
		if(RS422_RX_D(1) = '1') and (RS422_RX_D2(1) = '0') then
			-- new bit
			if(RS422_RX_D(2) = '1') then
				-- byte clock. currently receiving MSB (MSB transmitted first)
				BIT_COUNT <= "000";
			else
				BIT_COUNT <= BIT_COUNT + 1;
		  	end if;
	 	end if;
  	end if;
end process; 

-- Convert RS422_RX(3) data stream from serial to 8-bit parallel.
-- MSB is transmitted first.
-- Store in temporary storage so that the actual output changes 8 bits at a time, not bit by bit.
SERIAL_TO_PARALLEL_CONVERSION_001: process(CLK_IN1G)
begin
	if rising_edge(CLK_IN1G) then
		if(RS422_RX_D(1) = '1') and (RS422_RX_D2(1) = '0') then
			-- new bit
			DATA_BYTE(0) <= RS422_RX_D(3);
			DATA_BYTE(6 downto 1) <= DATA_BYTE(5 downto 0);	-- shift left
		end if;
  	end if;
end process; 

-- Latch out the data byte upon receiving bit 0 (LSB)
-- generate SAMPLE_CLK_OUT to tell the COM-8001 that data is ready
DATA_BYTE_OUT_001: process(CLK_IN1G)
begin
	if rising_edge(CLK_IN1G) then
		if(RS422_RX_D(1) = '1') and (RS422_RX_D2(1) = '0') and (BIT_COUNT = 7) then
			-- received last bit
			SAMPLE_CLK_OUT <= '1';
			DATA_BYTE_OUT(7 downto 1) <= DATA_BYTE(6 downto 0);
			DATA_BYTE_OUT(0) <= RS422_RX_D(3);
		else
			SAMPLE_CLK_OUT <= '0';
		end if;
  	end if;
end process; 

TRIGGER_OUT <= J6_A0_A1(5);

--// Connection matrix A4 -------------------------
-- Extend sample clock request from 25 ns to 50 ns pulse 
-- to meet the 10 MHz maximum frequency on RS422 
RECLOCK_005: process(CLK_IN1G, J6)
begin
	if rising_edge(CLK_IN1G) then
		J6_13_D <= J6(13);
		J6_13_D2 <= J6_13_D;
		J6_13_D3 <= J6_13_D2;
  	end if;
end process;



--// J6 output multiplexer
J6_OUTPUT_MUX_001: process(CPLD_CNTRL, J6_A0_A1, SAMPLE_CLK_OUT, DATA_BYTE_OUT)
begin
	case(CPLD_CNTRL) is
		when "0000" => 
			-- Connection matrices A0 
			J6(5 downto 2) <= J6_A0_A1(5 downto 2);
			J6(9 downto 6) <= (others => 'Z');
			J6(12 downto 11) <= (others => 'Z');
		when "0001" => 
			-- Connection matrices A1
			J6(5 downto 2) <= J6_A0_A1(5 downto 2);
			J6(9 downto 6) <= (others => 'Z');
			J6(12 downto 11) <= (others => 'Z');
		when "0010" => 
			-- Connection matrix A2
			J6(2) <= SAMPLE_CLK_OUT;
			J6(3) <= DATA_BYTE_OUT(0);
			J6(4) <= DATA_BYTE_OUT(1);
			J6(5) <= DATA_BYTE_OUT(2);
			J6(6) <= DATA_BYTE_OUT(3);
			J6(7) <= DATA_BYTE_OUT(4);
			J6(8) <= DATA_BYTE_OUT(5);
			J6(9) <= DATA_BYTE_OUT(6);
			J6(11) <= DATA_BYTE_OUT(7);
			J6(12) <= TRIGGER_OUT;
		when "0100" => 
			-- Connection matrix A4
			J6(2) <= J6_13_D3;
			J6(3) <= RS422_RX(1);
			J6(9 downto 4) <= (others => 'Z');
			J6(12 downto 11) <= (others => 'Z');
		when "0101" => 
			-- Connection matrix A5
			J6(2) <= J6_13_D3;
			J6(3) <= RS422_RX(1);
			J6(9 downto 4) <= (others => 'Z');
			J6(12 downto 11) <= (others => 'Z');
		when others => 
			J6(9 downto 2) <= (others => 'Z');
			J6(12 downto 11) <= (others => 'Z');
  	end case;
end process;

--// J2 output multiplexer
J2_OUTPUT_MUX_001: process(CPLD_CNTRL, RS422_RX)
begin
	case(CPLD_CNTRL) is
		when "0011" => 
			-- Connection matrix A3 
			J2(24 downto 21) <= RS422_RX(4 downto 1);
		when others => 
			J2(24 downto 21) <= (others => 'Z');
  	end case;
end process;



end Behavioral;
